#ifndef MIXSNET_MIWSNET_H
#define MIXSNET_MIWSNET_H
#include<ColoredWeightedGraph.hpp>
#include<Matrix.hpp>
#include<Algorithms.hpp>
#include<Functors.hpp>
#include<cluster.h>
#include<iostream>
#include<fstream>
#include<vector>
#include<utility> //<pair>
#include<limits>
#include<algorithm>
#include <utility>
#include <chrono>

namespace mixsnet{

  const unsigned int NBITERATIONFPMAX = 30;
  const double EPSILON = 0.0001;

  template<typename weight_type_t, typename data_type_t, class TDistribution>
  class MixSNet
  {
  protected:
    TDistribution& _dist;
    ColoredRoadNetwork<weight_type_t>& _net;
    unsigned int _nbcol;
    Matrix<data_type_t>* _Yptr;
    bool _directed;
    double _lambda;
    std::vector<int> _nalpha;
    std::vector<color_type> _curcolors;
    color_type getCurrentColor(const typename ColoredRoadNetwork<weight_type_t>::Vertex& v) const {return _curcolors[_net.getIndex(v)];}
    void setCurrentColor(const typename ColoredRoadNetwork<weight_type_t>::Vertex& v, color_type c) {_curcolors[_net.getIndex(v)]=c;}
    void setCurrentColors(const std::vector<color_type>& colors) {_curcolors=colors;}
    void saveColors();
    void loadColors();
    void reinit(unsigned int nbcol);
    void correctSingleton();
    void updateNAlpha();
    template<class TPenalty>
    std::pair<double,double> completeLogLikelihood();
    double colorsLogLikelihood();
    template<class TPenalty>
    bool fixedPoint();
    template<class TPenalty>
    double fixedPointCore(unsigned int i, const typename ColoredRoadNetwork<weight_type_t>::Vertex& vstar, color_type cold, color_type cnew);
    template<class TPenalty>
    void eStep(){
      fixedPoint<TPenalty>();
    }
    void mStep();
    void spatialSpreadingInitializer(bool random = false);
    void spatialSpreadingInitializerCore(color_type cinterest, color_type cnew, bool random = false);
    void noInitializer();
    void hclustInitializer();
  public:
    MixSNet(ColoredRoadNetwork<weight_type_t>& net, TDistribution& dist, bool directed = false);
    MixSNet(const MixSNet& model);
    virtual ~MixSNet() {}
    void setLambda(double lambda){ _lambda = lambda; }
    void initialize(Matrix<data_type_t>& Y, char initmethod = 'H');
    void initialize(Matrix<data_type_t>& Y, const std::vector<color_type>& colors);
    template<class TPenalty>
    std::pair<double,double> estimate(Matrix<data_type_t>& Y, unsigned int NBITERATIONEMMAX = 20);
    template<class TPenalty>
    double completeLogLikelihood(Matrix<data_type_t>& Y);
    template<class TPenalty>
    double penaltyTerm() {return TPenalty().completeLikelihoodTerm(_net, _curcolors);}
    const std::vector<color_type>& getCurrentColors()  const {return _curcolors;}
    Matrix<data_type_t> draw();
    template <typename weight_type_t1, typename data_type_t1, class TDistribution1>
    friend std::ostream& operator<<(std::ostream& out, MixSNet<weight_type_t1, data_type_t1, TDistribution1>& model);
    template <typename weight_type_t1, typename data_type_t1, class TDistribution1>
    friend std::istream&  operator>>( std::istream& in, MixSNet<weight_type_t1, data_type_t1, TDistribution1>& model);
  };

  template<typename weight_type_t, typename data_type_t, class TDistribution>
  MixSNet<weight_type_t, data_type_t, TDistribution>::MixSNet(ColoredRoadNetwork<weight_type_t>& net, TDistribution& dist, bool directed)
  : _dist(dist), _net(net), _nbcol(dist.nbColors()), _Yptr(NULL), _directed(directed), _lambda(0.), _nalpha(_nbcol, 0), _curcolors(_net.nbVertices(), color_default) {}

  template<typename weight_type_t, typename data_type_t, class TDistribution>
   MixSNet<weight_type_t, data_type_t, TDistribution>::MixSNet(const MixSNet<weight_type_t, data_type_t, TDistribution>& model)
   : _dist(model._dist), _net(model._net), _nbcol(model._nbcol), _Yptr(model._Yptr), _directed(model._directed), _lambda(model._lambda), _nalpha(model._nalpha), _curcolors(model._curcolors) {}

  template<typename weight_type_t, typename data_type_t, class TDistribution>
  void MixSNet<weight_type_t, data_type_t, TDistribution>::reinit(unsigned int nbcol){
    _nbcol = nbcol;
    _dist.reinit(nbcol);
    _nalpha.assign(nbcol, 0);
    _curcolors.assign(_curcolors.size(), color_default);
  }

  template<typename weight_type_t, typename data_type_t, class TDistribution>
  void MixSNet<weight_type_t, data_type_t, TDistribution>::initialize(Matrix<data_type_t>& Y, char initmethod)
  {
    _Yptr = &Y;
    reinit(_nbcol);
    switch (initmethod){
    case 'N' : noInitializer();break;
    case 'H' : hclustInitializer();break;
    case 'S' : spatialSpreadingInitializer();break;
    case 'R' : spatialSpreadingInitializer(true);break;
    }
  }

  template<typename weight_type_t, typename data_type_t, class TDistribution>
  void MixSNet<weight_type_t, data_type_t, TDistribution>::initialize(Matrix<data_type_t>& Y, const std::vector<color_type>& colors)
  {
    _Yptr = &Y;
    _curcolors=colors;
    saveColors();
    noInitializer();
  }

  template<typename weight_type_t, typename data_type_t, class TDistribution>
  template<class TPenalty>
  std::pair<double,double> MixSNet<weight_type_t, data_type_t, TDistribution>::estimate(Matrix<data_type_t>& Y, unsigned int NBITERATIONEMMAX)
  {
    _Yptr = &Y;
    srand ( time(NULL) + getpid() );
    bool end = false;
    std::pair<double,double> lklinfo = completeLogLikelihood<TPenalty>();
    double lkl1 = lklinfo.first, lkl2;
    for (unsigned int i=0; i<NBITERATIONEMMAX && !end; i++){
      //std::cerr<<"####### ESTEP ########"<<std::endl;
      eStep<TPenalty>();
      lklinfo = completeLogLikelihood<TPenalty>();
      lkl2 = lklinfo.first;
      //std::cerr<<"####### MSTEP ########"<<std::endl;
      mStep();
      lklinfo = completeLogLikelihood<TPenalty>();
      lkl2 = lklinfo.first;
      if (lkl2-lkl1<0){
        std::cerr<<"Warning: CompleteLikelihood is decreasing - Uncorrect behavior"<<std::endl;
        end = false;
      }
      else{
          end = bool((lkl2-lkl1)/(-lkl1) < EPSILON);
      }
      lkl1 = lkl2;
    }
    //std::cerr<<"Final: "<<_lambda<<" "<<lklinfo.first<<" "<<lklinfo.second<<std::endl;
    _Yptr = NULL;
    return lklinfo;
  }

  template<typename weight_type_t, typename data_type_t, class TDistribution>
  template<class TPenalty>
  bool MixSNet<weight_type_t, data_type_t, TDistribution>::fixedPoint()
  {
    // backup of vertex colors
    std::vector<color_type> newcolors = _curcolors;
    // fixed point loop
    bool maxreached = false;
    unsigned int nbiteration = 0;
    typename ColoredRoadNetwork<weight_type_t>::VertexIterator itv, itvend = _net.verticesEnd();
    while ((!maxreached) && (nbiteration<NBITERATIONFPMAX)){
      //std::cerr<<"=== Iteration "<<nbiteration<<std::endl;
        nbiteration++;
        maxreached = true;
        for(itv = _net.verticesBegin(); itv < itvend; itv++){ // < instead of != for openmp
            unsigned int i = _net.getIndex(*itv);
            double taumax = -std::numeric_limits<double>::max();
            color_type cold = getCurrentColor(*itv);
            color_type cmax = cold;
            for (color_type cnew = 0; cnew<_nbcol; cnew++){
                double tau = fixedPointCore<TPenalty>(i, *itv, cold, cnew);
                //std::cerr<<"FP "<<tau<<std::endl;
                if (tau>taumax){
                    taumax = tau;
                    cmax = cnew;
                }
            }
            // changing color
            //std::cerr<<"Summary: for "<<_net.getName(*itv)<<" we have taumax="<<taumax<<" to switch from "<<cold<<" to "<<cmax<<std::endl;
            if (cold != cmax){
	      //std::cerr<<"Moving "<<_net.getName(*itv)<<" from Color "<<cold<<" to Color "<<cmax<<std::endl;
                maxreached = false;
            }
            newcolors[i] = cmax;
        }
        if (!maxreached){
            itv = _net.verticesBegin();
            _curcolors = newcolors;
        }
    }
    //std::cerr<<"======== EStep Color "<<std::endl;
    itv = _net.verticesBegin();
    //for(unsigned int i=0; itv != itvend; itv++, i++){
    //std::cerr<<"Color "<<getCurrentColor(*itv)<<" for "<<_net.getName(*itv)<<std::endl;
    //}
    return maxreached;
  }

  template<typename weight_type_t, typename data_type_t, class TDistribution>
  template<class TPenalty>
  double MixSNet<weight_type_t, data_type_t, TDistribution>::fixedPointCore(unsigned int i, const typename ColoredRoadNetwork<weight_type_t>::Vertex& vstar, color_type cold, color_type cnew)
  {
    setCurrentColor(vstar, cnew);
    double tau1 = forEachYidot<double,
        ColoredRoadNetwork<weight_type_t>,
        Matrix<data_type_t>,
        CompleteLogLikelihoodFunctor<data_type_t, TDistribution>,
        decltype(_curcolors)>
    (i, _net, *this->_Yptr, CompleteLogLikelihoodFunctor<data_type_t, TDistribution>(this->_dist), _curcolors, _directed);
    double tau2 = log(double(_nalpha[cnew])/double(_net.nbVertices()));
    double tau3 = _lambda * TPenalty().fixedPointTerm(i, _net, _curcolors);
    //std::cerr<<_lambda<<" "<<TPenalty().fixedPointTerm(i, _net, _curcolors)<<" = "<<tau3<<std::endl;
    setCurrentColor(vstar, cold);
    return tau1 + tau2 - tau3;
  }

  template<typename weight_type_t, typename data_type_t, class TDistribution>
  void MixSNet<weight_type_t, data_type_t, TDistribution>::mStep()
  {
    updateNAlpha();
    //correctSingleton();
    saveColors();
    _dist.compute(_net, *_Yptr);
  }

  template<typename weight_type_t, typename data_type_t, class TDistribution>
  void MixSNet<weight_type_t, data_type_t, TDistribution>::correctSingleton()
  {
    for (color_type q=0; q<_nbcol; q++){
        if (_nalpha[q] == 1){
            typename ColoredRoadNetwork<weight_type_t>::VertexIterator itv = _net.verticesBegin();
            while (getCurrentColor(*itv) != q) itv++;
            // selecting nearest neighbor
            typename ColoredRoadNetwork<weight_type_t>::OutEdgeIterator ite = _net.outEdgesBegin(*itv), iteend = _net.outEdgesEnd(*itv);
            typename ColoredRoadNetwork<weight_type_t>::Vertex nearestv = target(*ite, _net);
            weight_type_t wmin = _net.getWeight(*ite);
            ite++;
            for (; ite != iteend; ite++){
                weight_type_t w = _net.getWeight(*ite);
                if (w<wmin){
                    wmin = w;
                    nearestv = target(*ite, _net);
                }
            }
            color_type cold = getCurrentColor(nearestv), cnew = q;
            if (_nalpha[getCurrentColor(nearestv)]>2){
                setCurrentColor(nearestv, cnew);
                _nalpha[cold]--;
                _nalpha[cnew]++;
                //std::cerr<<"Singleton case corrected"<<std::endl;
                //std::cerr<<"Color "<<getCurrentColor(nearestv)<<" for "<<_net.getName(nearestv)<<std::endl;
            }
            else{
                std::cerr<<"Error: enable to correct singleton case"<<std::endl;
                exit(1);
            }

        }
    }
  }

  template<typename weight_type_t, typename data_type_t, class TDistribution>
  void MixSNet<weight_type_t, data_type_t, TDistribution>::updateNAlpha()
  {
    //std::cerr<<"updateNAlpha"<<std::endl;
    _nalpha.assign(_nbcol, 0);
    typename ColoredRoadNetwork<weight_type_t>::VertexIterator it = _net.verticesBegin();
    unsigned int nbv = _net.nbVertices();
    for (unsigned int k=0; k<nbv; k++){_nalpha[getCurrentColor(*it)]++; it++;}
  }

  template<typename weight_type_t, typename data_type_t, class TDistribution>
  template<class TPenalty>
  double MixSNet<weight_type_t, data_type_t, TDistribution>::completeLogLikelihood(Matrix<data_type_t>& Y)
  {
    _Yptr = &Y;
    double lkl = completeLogLikelihood<TPenalty>().first;
    _Yptr = NULL;
    return lkl;
  }

  template<typename weight_type_t, typename data_type_t, class TDistribution>
  template<class TPenalty>
  std::pair<double,double> MixSNet<weight_type_t, data_type_t, TDistribution>::completeLogLikelihood()
  {
    // log(P(Y|Z,theta)
    double l1 = forEachYij<double,
        ColoredRoadNetwork<weight_type_t>,
        Matrix<data_type_t>,
        CompleteLogLikelihoodFunctor<data_type_t, TDistribution>,
        decltype(_curcolors)>
      (_net, *_Yptr, CompleteLogLikelihoodFunctor<data_type_t, TDistribution>(_dist), _curcolors, _directed);
    /*
    double ltmp = 0;
    for (unsigned int k=0; k<_net.nbVertices(); k++)
      ltmp += forEachYidot<double,
      ColoredRoadNetwork<weight_type_t>,
      Matrix<data_type_t>,
      CompleteLogLikelihoodFunctor<data_type_t, TDistribution>,
      decltype(_curcolors)>
    (k, _net, *this->_Yptr, CompleteLogLikelihoodFunctor<data_type_t, TDistribution>(this->_dist), _curcolors, _directed);

    std::cout<<"l1: "<<l1<<" ltmp: "<<ltmp<<" directed: "<<_directed<<std::endl;
    */

    // log(P(Z,alpha))
    double l2 = colorsLogLikelihood();
    double l3 = _lambda * TPenalty().completeLikelihoodTerm(_net, _curcolors);
    std::pair<double,double> res;
    res.first = l1 + l2 - l3;
    res.second = l3;
    return res;
  }

  template<typename weight_type_t, typename data_type_t, class TDistribution>
  double MixSNet<weight_type_t, data_type_t, TDistribution>::colorsLogLikelihood()
  {
    // log(P(Z,alpha))
    double l2 = 0.;
    for (color_type c = 0; c<_nbcol; c++){
        int nalpha = _nalpha[c];
        if (nalpha>0)
            l2 += log(double(nalpha)/double(_net.nbVertices()));
    }
    return l2;
  }

  template<typename weight_type_t, typename data_type_t, class TDistribution>
  void MixSNet<weight_type_t, data_type_t, TDistribution>::spatialSpreadingInitializerCore(color_type cinterest, color_type cnew, bool random)
  {
    unsigned int i=0, j=0;
    typename ColoredRoadNetwork<weight_type_t>::VertexIterator itv;
    unsigned int nbv = _net.nbVertices();
    if (random){
        //  finding two anchors at random vertices of color cinterest
        unsigned int ii,jj;
        if (_nalpha[cinterest]<=1){
            std::cerr<<"Error in SpatialSpreading : a singleton can not be split."<<std::endl;
            exit(1);
        }
        std::uniform_int_distribution<> uid(0,_nalpha[cinterest]-1);
        unsigned seed = std::chrono::system_clock::now().time_since_epoch().count();
        std::default_random_engine dre(seed);
	ii = uid(dre); jj = ii;  while(ii==jj) jj = uid(dre);
        //std::cout<<"II="<<ii<<" JJ="<<jj<<std::endl;
        unsigned int kk = 0;
        itv = _net.verticesBegin();
        for (unsigned int k=0; k<nbv; k++){
            if (getCurrentColor(*itv) == cinterest){
                if (kk == ii) i = k;
                if (kk == jj) j = k;
                kk++;
            }
            itv++;
        }
        //std::cerr<<"Random couple is :"<<_net.getName(_net.getVertexByIndex(i))<<" "
        //<<_net.getName(_net.getVertexByIndex(j))<<std::endl;
    }
    else{
        //  finding two anchors as farthest vertices of color cinterest
        std::pair<unsigned int, unsigned int> farthest = forEachYij<std::pair<unsigned int, unsigned int>,
            ColoredRoadNetwork<weight_type_t>,
            Matrix<data_type_t>,
            SpatialSpreadingFunctor<weight_type_t, data_type_t>,
            decltype(_curcolors)>
        (_net, *_Yptr, SpatialSpreadingFunctor<weight_type_t, data_type_t>(_net, cinterest), _curcolors, _directed);
        i = farthest.first, j = farthest.second;
        //std::cerr<<"Farthest couple is :"<<_net.getName(_net.getVertexByIndex(i))<<" "
        //<<_net.getName(_net.getVertexByIndex(j))<<std::endl;
    }
    // assigning vertices to nearest anchor
    itv = _net.verticesBegin();
    for (unsigned int k=0; k<nbv; k++){
        if (getCurrentColor(*itv) == cinterest){
            //std::cerr<<"Shortest between "<<_net.shortestPath(k, i)<<" and "<<_net.shortestPath(k, j)<<" ?"<<std::endl;
            if (_net.shortestPath(k, i) > _net.shortestPath(k, j))
              setCurrentColor(*itv, cinterest);
            else
              setCurrentColor(*itv, cnew);
            //std::cerr<<"Assigning "<<getCurrentColor(*itv)<<" to "<<_net.getName(*itv)<<std::endl;
        }
        itv++;
    }
  }

  template<typename weight_type_t, typename data_type_t, class TDistribution>
  void MixSNet<weight_type_t, data_type_t, TDistribution>::spatialSpreadingInitializer(bool random)
  {
    _nalpha[color_default] = _net.nbVertices();
    color_type cinterest = color_default;
    for (color_type cnew=1; cnew<_nbcol; cnew++){
        spatialSpreadingInitializerCore(cinterest, cnew, random);
        updateNAlpha();
        cinterest = std::max_element(_nalpha.begin(), _nalpha.end()) - _nalpha.begin(); // color with larger alpha
    }
    typename ColoredRoadNetwork<weight_type_t>::VertexIterator itv = _net.verticesBegin(), itvend = _net.verticesEnd();
    //for(; itv != itvend; itv++) std::cerr<<"Color "<<getCurrentColor(*itv)<<" for "<<_net.getName(*itv)<<std::endl;
    mStep();
  }

  template<typename weight_type_t, typename data_type_t, class TDistribution>
  void MixSNet<weight_type_t, data_type_t, TDistribution>::noInitializer()
  {
    loadColors();
    typename ColoredRoadNetwork<weight_type_t>::VertexIterator itv = _net.verticesBegin(), itvend = _net.verticesEnd();
    for(unsigned int i=0; itv != itvend; itv++, i++){
      //std::cerr<<"Color "<<getCurrentColor(*itv)<<" for "<<_net.getName(*itv)<<std::endl;
    }
    mStep();
  }

  template<typename weight_type_t, typename data_type_t, class TDistribution>
  void MixSNet<weight_type_t, data_type_t, TDistribution>::hclustInitializer()
  {
    /*
    unsigned int nbv = _net.nbVertices();
    double** distmatrix = new double*[nbv];
    distmatrix[0] = new double[nbv*nbv];
    for (unsigned int i=1; i<nbv; i++) distmatrix[i] = distmatrix[i-1]+nbv;
    for (unsigned int i=0; i<nbv; i++)
      for (unsigned int j=0; j<nbv; j++){
          if (i==j)
            distmatrix[i][i] = 0.;
          else{
              data_type_t yij = (*_Yptr)(i,j);
              if (yij>0)
                distmatrix[i][j] = double(yij);
              else{
                  std::cerr<<"Warning: negative value "<<yij<<" in Y could be source of instability"<<std::endl;
                  distmatrix[i][j] = 0.;
              }
          }
      }
    Node* tree =  treecluster(nbv, nbv, NULL, NULL, NULL, -1, 'e', 'm', distmatrix);
    delete[] distmatrix[0]; delete[] distmatrix;
    int clusterid[nbv];
    cuttree(nbv, tree, _nbcol, clusterid);
    free(tree);
    // saving as colors
    typename ColoredRoadNetwork<weight_type_t>::VertexIterator itv = _net.verticesBegin(), itvend = _net.verticesEnd();
    for(unsigned int i=0; itv != itvend; itv++, i++){
        setCurrentColor(*itv, color_type(clusterid[i]));
        //std::cerr<<"Color "<<getCurrentColor(*itv)<<" for "<<_net.getName(*itv)<<std::endl;
    }
    itv = _net.verticesBegin(), itvend = _net.verticesEnd();
    for(unsigned int i=0; itv != itvend; itv++, i++){
    	std::cerr<<"Color "<<getCurrentColor(*itv)<<" for "<<_net.getName(*itv)<<std::endl;
    }

    mStep();
    */
  }

  template<typename weight_type_t, typename data_type_t, class TDistribution>
  Matrix<data_type_t> MixSNet<weight_type_t, data_type_t, TDistribution>::draw()
  {
    unsigned int nbv = _net.nbVertices();
    Matrix<data_type_t> Y(nbv);
    forEachYij<void,
      ColoredRoadNetwork<weight_type_t>,
      Matrix<data_type_t>,
      DrawFunctor<data_type_t, TDistribution>,
      decltype(_curcolors)>
    (_net, Y, DrawFunctor<data_type_t, TDistribution>(_dist), _curcolors, _directed);
    // symmetrizing
    if (!_directed)
      for (unsigned int i=0; i<nbv; i++)
        for (unsigned int j=i+1; j<nbv; j++)
          Y(i,j) = Y(j,i);
//    for (unsigned int i=0; i<nbv; i++){
//        std::cerr<<_net.getName(_net.getVertexByIndex(i))<<" ";
//        for (unsigned int j=0; j<i; j++)
//          std::cerr<<Y(i,j)<<" ";
//        std::cerr<<"0 ";
//        for (unsigned int j=i+1; j<nbv; j++)
//          std::cerr<<Y(j,i)<<" ";
//        std::cerr<<std::endl;
//    }
    return Y;
  }

  template <typename weight_type_t, typename data_type_t, class TDistribution>
  std::ostream& operator<<(std::ostream& out, MixSNet<weight_type_t, data_type_t, TDistribution>& model)
  {
    out<<"# Q\n"<<model._nbcol<<std::endl;
    out<<"# Alpha\n";
    for (color_type q=0; q<model._nbcol; q++)
      out<<double(model._nalpha[q])/double(model._net.nbVertices())<<" ";
    out<<std::endl;
    out<<model._dist;
    model.saveColors();
    out<<model._net;
    return out;
  }

  template <typename weight_type_t, typename data_type_t, class TDistribution>
  std::istream& operator>>(std::istream& in, MixSNet<weight_type_t, data_type_t, TDistribution>& model)
  {
    std::string line;
    // # Q
    getline(in,line);
    getline(in,line); std::istringstream linestream1(line);
    unsigned int nbcol; linestream1>>nbcol;
    if (model._nbcol != nbcol){
        model.reinit(nbcol);
    }
    line.clear();
    // # Alpha
    getline(in,line);
    getline(in,line); std::istringstream linestream2(line);
    for (color_type q=0; q<model._nbcol; q++){
       double x; linestream2>>x;
       model._nalpha[q] = floor(x*model._net.nbVertices()+0.5);
    }
    in>>model._dist;
    in>>model._net;
    model.loadColors();

    return in;
  }

  template<typename weight_type_t, typename data_type_t, class TDistribution>
  void MixSNet<weight_type_t, data_type_t, TDistribution>::saveColors()
  {
      typename ColoredRoadNetwork<weight_type_t>::VertexIterator itv = _net.verticesBegin(), itvend = _net.verticesEnd();
      for(unsigned int i=0; itv != itvend; itv++, i++)
        _net.setColor(*itv, getCurrentColor(*itv));
  }

  template<typename weight_type_t, typename data_type_t, class TDistribution>
  void MixSNet<weight_type_t, data_type_t, TDistribution>::loadColors()
  {
      typename ColoredRoadNetwork<weight_type_t>::VertexIterator itv = _net.verticesBegin(), itvend = _net.verticesEnd();
      for(unsigned int i=0; itv != itvend; itv++, i++)
        setCurrentColor(*itv, _net.getColor(*itv));
  }

}
#endif
