#ifndef MIXSNET_PENALTY_H
#define MIXSNET_PENALTY_H
#include<iostream>
namespace mixsnet{

  template<typename weight_type_t>
  class NoPenalty
  {
  public:
    NoPenalty() {}
    ~NoPenalty() {}
    double completeLikelihoodTerm(ColoredRoadNetwork<weight_type_t>& net, std::vector<color_type>& colors) {return 0.;}
    double fixedPointTerm(unsigned int i, ColoredRoadNetwork<weight_type_t>& net, std::vector<color_type>& colors) {return 0.;}
  };

  template<typename weight_type_t, class SijComputer>
  class NeighborhoodPenalty
  {
  public:
    NeighborhoodPenalty() {}
    virtual ~NeighborhoodPenalty() {}
    inline double completeLikelihoodTerm(ColoredRoadNetwork<weight_type_t>& net, std::vector<color_type>& colors);
    inline double fixedPointTerm(unsigned int i, ColoredRoadNetwork<weight_type_t>& net, std::vector<color_type>& colors);
  };

  template <typename weight_type_t, class SijComputer>
  double NeighborhoodPenalty<weight_type_t, SijComputer>::completeLikelihoodTerm(ColoredRoadNetwork<weight_type_t>& net, std::vector<color_type>& colors){
    typename ColoredRoadNetwork<weight_type_t>::VertexIterator itv = net.verticesBegin();
    unsigned int nbv = net.nbVertices();
    double pen = 0.;
    for (unsigned int i=0; i<nbv; i++){
        pen += fixedPointTerm(i, net, colors);
        itv++;
    }
    return 0.5*pen;
  }

  template <typename weight_type_t, class SijComputer>
  double NeighborhoodPenalty<weight_type_t, SijComputer>::fixedPointTerm(unsigned int i, ColoredRoadNetwork<weight_type_t>& net, std::vector<color_type>& colors){
    typename ColoredRoadNetwork<weight_type_t>::Vertex v = net.getVertexByIndex(i), v2;
    color_type c = colors[i], c2;
    typename ColoredRoadNetwork<weight_type_t>::OutEdgeIterator ite = net.outEdgesBegin(v), iteend = net.outEdgesEnd(v);
    double pen = 0.;
    for (; ite != iteend; ite++){
        v2 = target(*ite, net);
        c2 = colors[net.getIndex(v2)];
        if (c!=c2){
            weight_type_t xij = net.getWeight(*ite);
            double sij = double(SijComputer()(net, v, xij));
            pen += sij;
            //std::cerr<<sij<<std::endl;
        }
        //std::cerr<<i<<" "<<c<<" "<<c2<<std::endl;
    }
    return 2*pen;
  }

  template<typename weight_type_t>
  class SijProximity
  {
  public:
    SijProximity() {}
    ~SijProximity() {}
    weight_type_t operator()(ColoredRoadNetwork<weight_type_t>& net, typename ColoredRoadNetwork<weight_type_t>::Vertex& v, const weight_type_t& xij){
      return net.getMaxWeight() - xij + 1;
    }
  };

  template<typename weight_type_t>
  class SijProximityDegreeNormalized
  {
  public:
    SijProximityDegreeNormalized() {}
    ~SijProximityDegreeNormalized() {}
    double operator()(ColoredRoadNetwork<weight_type_t>& net, typename ColoredRoadNetwork<weight_type_t>::Vertex& v, const weight_type_t& xij){
      return double(SijProximity<weight_type_t>()(net,v,xij)) / double(degree(v, net));
    }
  };

  template<typename weight_type_t>
  class SijLaplacian
  {
  public:
    SijLaplacian() {}
    ~SijLaplacian() {}
    weight_type_t operator()(ColoredRoadNetwork<weight_type_t>& net, typename ColoredRoadNetwork<weight_type_t>::Vertex& v, const weight_type_t& xij){
      return xij;
    }
  };

  template<typename weight_type_t>
  class SijLaplacianDegreeNormalized
  {
  public:
    SijLaplacianDegreeNormalized() {}
    ~SijLaplacianDegreeNormalized() {}
    double operator()(ColoredRoadNetwork<weight_type_t>& net, typename ColoredRoadNetwork<weight_type_t>::Vertex& v, const weight_type_t& xij){
      return double(SijLaplacian<weight_type_t>()(net, v, xij)) / double(degree(v, net));
    }
  };

}
#endif
