#include<MixSNet.hpp>
#include<OMPConductor.hpp>
#include<GaussianDistribution.hpp>
#include<Penalty.hpp>
#include<string>
#include<iostream>
#include<limits>
#include<fstream>
#include<vector>
#include<stdlib.h>
#include<boost/program_options.hpp>
using namespace std;
using namespace mixsnet;
namespace po = boost::program_options;

int main(int argc, char **argv)
{
  // BEGIN PROGRAM OPTIONS //
   // Declare a group of options that will be allowed only on command line
   po::options_description generic("Generic options");
   generic.add_options()
     ("help,h", "Display this information")
     ("version,v", "Display version information");
   // Declare a group of options that will be allowed both on command line and in config file
   po::options_description config("Configuration");
   string initmethods, initfile;
   unsigned int nbcolors, burnnbiteration, randomnbiteration;
   double increment, ldepth, clump;
   config.add_options()
     ("method,m", po::value<string>(&initmethods)->default_value("H","H"),"Init method(s) token composed with {H,S,R,N} for Hclust, Spatial, Random and Noinit")
     ("nbcolors,q", po::value<unsigned int>(&nbcolors)->default_value(2,"2"), "Number of groups")
     ("ldepth,d", po::value<double>(&ldepth)->default_value(3.,"3."), "Lambda depth as a factor of the complete likelihood")
     ("increment,i", po::value<double>(&increment)->default_value(0.5,"0.5"), "Lambda increment")
     ("clump,c", po::value<double>(&clump)->default_value(-std::numeric_limits<double>::max(),"NA"), "Clump value in case of inflated distribution")
     ("nbburn,b", po::value<unsigned int>(&burnnbiteration)->default_value(0,"0"), "Number of iterations in the burn step")
     ("nbrand,r", po::value<unsigned int>(&randomnbiteration)->default_value(0,"100"), "Number of random initializations (if init method is R)")
     ("weighted,w", "Specifying the network is weighted")
     ("groups,g", po::value<string>(&initfile)->default_value(""),"File containing initial groups to be loaded (\"#Colors\" on first line) - Only with \"-m N\"");
   // Hidden options, will be allowed both on command line and in config file, but will not be shown to the user.
   po::options_description hidden("Hidden options");
   hidden.add_options()
     ("input-file", po::value< vector<string> >(), "input file");

   po::options_description cmdline_options;
   cmdline_options.add(generic).add(config).add(hidden);
   po::options_description visible("");
   visible.add(generic).add(config);
   po::positional_options_description pd;
   pd.add("input-file", 2);
   po::variables_map vm;
   store(po::command_line_parser(argc, argv).
         options(cmdline_options).positional(pd).run(), vm);
   po::notify(vm);
   // HELP
   if (vm.count("help")) {
     stringstream infostream;
     infostream<<"geomix [OPTIONS] <NETFILE> <DATAFILE>"<<endl<<endl;
     cout<<"geomix, do something "<<endl;
     cout<<"General use:"<<endl<<"  "<<infostream.str()<<endl;
     cout << visible << "\n";
     return 0;
   }
   if (vm.count("version")) {
     cout<<"geomix, do something -- ";
     stringstream infostream;
     string VERSION("TODO TODO");
     infostream<<"version "<<VERSION<<endl;
     cout<<infostream.str()<<endl;
     return 0;
   }
   // Formatting
   bool badinput = false;
   if (!vm.count("input-file"))
     badinput = true;
   else if (vm["input-file"].as< vector<string> >().size()!=2)
     badinput = true;
   if (badinput){
     cerr<<"error: uncorrect number of command line arguments"<<endl;
     return 1;
   }
   bool isweighted = false;
   if (vm.count("weighted"))
     isweighted = true;
   vector<string> inputfiles = vm["input-file"].as< vector<string> >();
   // END PROGRAM OPTIONS //

   ColoredRoadNetwork<double> net(inputfiles[0]);
   Matrix<double> mat;
   mat.read(inputfiles[1], net.getConvertToIndexFunction());

   if (clump > -std::numeric_limits<double>::max()){
       InflatedDistribution<HomoscedasticGaussian> dist(nbcolors, clump);
       OMPConductor<double, double, InflatedDistribution<HomoscedasticGaussian> > runner;
       if ((initmethods.find('N') != std::string::npos) && (initfile.size())){
         runner.activateInput();
         ifstream finit(initfile.c_str());
         finit>>net;
         finit.close();
       }
       if (initmethods.find('S') != std::string::npos)
         runner.activateSpatial();
       if (initmethods.find('R') != std::string::npos)
         runner.activateRandom(randomnbiteration);
       runner.setBurnNbIteration(burnnbiteration);
       if (isweighted)
         runner.pipeline<NeighborhoodPenalty<double, SijProximity<double> > >(net, dist, mat, ldepth, increment);
       else
         runner.pipeline<NeighborhoodPenalty<double, SijLaplacian<double> > >(net, dist, mat, ldepth, increment);
   }
   else{
       HomoscedasticGaussian dist(nbcolors);
       OMPConductor<double, double, HomoscedasticGaussian> runner;
       if ((initmethods.find('N') != std::string::npos) && (initfile.size())){
         runner.activateInput();
         ifstream finit(initfile.c_str());
         finit>>net;
         finit.close();
       }
       if (initmethods.find('H') != std::string::npos)
         runner.activateHclust();
       if (initmethods.find('S') != std::string::npos)
         runner.activateSpatial();
       if (initmethods.find('R') != std::string::npos)
         runner.activateRandom(randomnbiteration);
       runner.setBurnNbIteration(burnnbiteration);
       if (isweighted)
         runner.pipeline<NeighborhoodPenalty<double, SijProximity<double> > >(net, dist, mat, ldepth, increment);
       else
         runner.pipeline<NeighborhoodPenalty<double, SijLaplacian<double> > >(net, dist, mat, ldepth, increment);
   }
   return 0;
}
