#include<MixSNet.hpp>
#include<OMPConductor.hpp>
#include<Bernouilli3DDistribution.hpp>
#include<Penalty.hpp>
#include<string>
#include<iostream>
#include<limits>
#include<fstream>
#include<vector>
#include<stdlib.h>
#include<boost/program_options.hpp>
using namespace std;
using namespace mixsnet;
namespace po = boost::program_options;

int main(int argc, char **argv)
{
  // BEGIN PROGRAM OPTIONS //
   // Declare a group of options that will be allowed only on command line
   po::options_description generic("Generic options");
   generic.add_options()
     ("help,h", "Display this information")
     ("version,v", "Display version information");
   // Declare a group of options that will be allowed both on command line and in config file
   po::options_description config("Configuration");
   string initfile;
   unsigned int nbcolors, randomnbiteration;
   config.add_options()
     ("nbcolors,q", po::value<unsigned int>(&nbcolors)->default_value(2,"2"), "Number of groups")
     ("nbrand,r", po::value<unsigned int>(&randomnbiteration)->default_value(0,"100"), "Number of random initializations")
     ("groups,g", po::value<string>(&initfile)->default_value(""),"File containing initial groups to be loaded (\"#Colors\" on first line) - Only with \"-m N\"");
   // Hidden options, will be allowed both on command line and in config file, but will not be shown to the user.
   po::options_description hidden("Hidden options");
   hidden.add_options()
     ("input-file", po::value< vector<string> >(), "input file");

   po::options_description cmdline_options;
   cmdline_options.add(generic).add(config).add(hidden);
   po::options_description visible("");
   visible.add(generic).add(config);
   po::positional_options_description pd;
   pd.add("input-file", 1);
   po::variables_map vm;
   store(po::command_line_parser(argc, argv).
         options(cmdline_options).positional(pd).run(), vm);
   po::notify(vm);
   // HELP
   if (vm.count("help")) {
     stringstream infostream;
     infostream<<"multiplexmix [OPTIONS]> <DATAFILE>"<<endl<<endl;
     cout<<"General use:"<<endl<<"  "<<infostream.str()<<endl;
     cout << visible << "\n";
     return 0;
   }
   if (vm.count("version")) {
     cout<<"multiplexmix - ";
     stringstream infostream;
     string VERSION("0.1.0");
     infostream<<"version "<<VERSION<<endl;
     cout<<infostream.str()<<endl;
     return 0;
   }
   // Formatting
   bool badinput = false;
   if (!vm.count("input-file"))
     badinput = true;
   else if (vm["input-file"].as< vector<string> >().size()!=1)
     badinput = true;
   if (badinput){
     cerr<<"error: uncorrect number of command line arguments"<<endl;
     return 1;
   }
   vector<string> inputfiles = vm["input-file"].as< vector<string> >();
   // END PROGRAM OPTIONS //



   EmptyColoredRoadNetwork net(inputfiles[0]);
   Matrix<Triplet> mat;
   mat.read(inputfiles[0], net.getConvertToIndexFunction());

   Bernouilli3D dist(nbcolors);
   OMPConductor<double, Triplet, Bernouilli3D> runner;
   runner.activateInput();
   ifstream finit(initfile.c_str());
   finit>>net;
   finit.close();
   runner.setBurnNbIteration(0);
   runner.pipeline<NoPenalty<double> >(net, dist, mat, 0., 1.);

   return 0;
}
